# Copyright (c) 2025 foofaraw (GitHub: foofaraw)
# Licensed under the MIT License (see LICENSE file for details).


class PromptError(Exception):
    """Base exception class for prompt-related errors."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class ConcatenationError(PromptError):
    """Exception raised for errors in the input prompt."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class ReplacementError(PromptError):
    """Exception class for replacement-related errors."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class ResponseError(PromptError):
    """Base exception class for response-related errors."""

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class MalformedOutput(ResponseError):
    """Exception raised for malformed outputs in responses."""

    def __init__(self, message: str) -> None:
        super().__init__(message)



# Copyright (c) 2025 foofaraw (GitHub: foofaraw)
# Licensed under the MIT License (see LICENSE file for details).

from __future__ import (  # Required for forward references in older Python versions
    annotations,
)

import copy
import inspect
import warnings
from typing import Any, Dict, List

from jinja2 import Environment, StrictUndefined, TemplateSyntaxError, nodes
from jinja2.visitor import NodeVisitor
from litellm import completion

from .exceptions import ConcatenationError, PromptError, ReplacementError, ResponseError
from .output import Output, outputs_to_xml
from .response import Response

_jinja_env = Environment(undefined=StrictUndefined)


def _extract_jinja_variables(template_source: str) -> List[str]:
    class OrderedVariableCollector(NodeVisitor):
        def __init__(self) -> None:
            self.variables: List[str] = []

        def visit_Name(self, node: nodes.Name) -> None:
            # The attribute 'name' may not be recognized by mypy on jinja2.nodes.Name.
            if node.name not in self.variables:
                self.variables.append(node.name)
            # generic_visit may be untyped.
            self.generic_visit(node)

        def visit_For(self, node: nodes.For) -> None:
            # Only visit the 'iter' part. The attribute 'iter' might not be recognized.
            self.visit(node.iter)
            # Skip visiting node.target, node.body, and node.else_ to avoid loop-local variables.

    env = Environment()
    parsed_template = env.parse(template_source)
    collector = OrderedVariableCollector()
    collector.visit(parsed_template)
    return collector.variables


def _jinja_substitution(template: str, **kwargs: Any) -> str:
    """
    Substitutes variables into a Jinja2 template string and returns the rendered result.

    :param template: str - The Jinja2 template string.
    :param kwargs: dict - Variables to substitute into the template.
    :return: str - The rendered template with variables substituted.
    """
    return _jinja_env.from_string(template).render(**kwargs)


def litellm_completion(prompt: str, *args: List[Any], **kwargs: Dict[str, Any]) -> str:
    response = completion(messages=[dict(role="user", content=prompt)], *args, **kwargs)
    return str(response.choices[0].message.content)


class Prompt:
    """A class representing a prompt template with support for variables and outputs.

    Attributes:
        template (str): The template string containing Jinja variables.
        outputs (List[Output]): List of outputs.
    """

    def __init__(self, template: str, returns: List[Output] = []):
        """Initialize a Prompt instance.

        Args:
            template (str): The template string.
            returns (List[Output]): List of outputs
        """
        self.template: str = template
        self.outputs: List[Output] = returns

        # Make sure there are no outputs with duplicate names
        errors = []
        index_for_name = dict()
        for index, output in enumerate(returns):
            name = output.name
            if name not in index_for_name:
                index_for_name[name] = index
            else:
                errors += [
                    f"Found outputs at positions {index_for_name[name]}, {index} with same name: '{name}'"
                ]
        if errors:
            raise PromptError("\n".join(errors))

    def __call__(self, *args: Any, **kwargs: Any) -> "Prompt":
        """Render the prompt with the given keyword arguments.

        Args:
            *args: Variables to substitute into the template.
            **kwargs: Named variables to substitute into the template.

        Returns:
            Prompt: A new Prompt instance with substituted template.
        """
        variable_names = self.get_variables()

        if len(args) > len(variable_names):
            raise ReplacementError(
                f"Provided {len(args)} positional arguments, but prompt template has only {len(variable_names)} variables"
            )

        for k, v in zip(variable_names, args):

            if k in kwargs:
                raise ReplacementError(
                    f"Got multiple values for template variable '{k}'"
                )
            kwargs[k] = v

        return Prompt(
            _jinja_substitution(self.template, **kwargs), returns=self.outputs
        )

    def __add__(self, other: Any) -> "Prompt":
        """Concatenate two prompts.

        Args:
            other (Union[str, Prompt]): The prompt or string to concatenate.

        Returns:
            Prompt: A new Prompt instance with combined template and outputs.

        Raises:
            PromptConcatenationError: If trying to add a non-string or non-Prompt object.
        """
        if isinstance(other, str):
            other_prompt = Prompt(other)
        elif isinstance(other, Prompt):
            other_prompt = other
        else:
            raise ConcatenationError(
                "Prompt error: trying to add Prompt to object other than str|Prompt for __add__"
            )

        return Prompt(
            self.template + "\n" + other_prompt.template,
            returns=self.outputs + other_prompt.outputs,
        )

    def __str__(self) -> str:
        """Render the prompt as a string, checking for undefined variables.

        Returns:
            str: The rendered prompt string.
        """

        if variables := self.get_variables():
            warning = f"Tried to render prompt that still has undefined Jinja2 variables: ({', '.join(sorted(variables))})"
            warnings.warn(warning, RuntimeWarning)

        string = self.template

        if self.outputs:
            outputs_with_content_indications = copy.deepcopy(self.outputs)
            for output in outputs_with_content_indications:
                if output.name:
                    output.content = f"... value for output {output.name} goes here ..."
                else:
                    output.content = "... value for this output goes here ..."

            string += (
                "\nProvide your response inside an XML such as this:\n"
                + outputs_to_xml(outputs_with_content_indications)
            )

        return string

    @classmethod
    def load(cls, template_file: str) -> "Prompt":
        """Load a prompt template from a file.

        Args:
            template_file (str): Path to the template file.

        Returns:
            Prompt: A new Prompt instance with the template content.
        """
        with open(template_file, "r") as file:
            return cls(file.read())

    def get_variables(self) -> List[str]:
        """Extract variables from the template.

        Returns:
            set[Any]: List of variable names present in the template.
        """

        try:
            return _extract_jinja_variables(self.template)
        except TemplateSyntaxError as e:
            warning = f"Tried to render prompt that contains incorrect Jinja2 template syntax ({str(e)})"
            warnings.warn(warning, RuntimeWarning)
            return []

    def returns(self, *args: Any, **kwargs: Any) -> "Prompt":
        """Add an output to the prompt.

        Args:
            *args: inputs for the Output constructor
            **kwargs: inputs for the Output constructor

        Returns:
            Prompt: A new Prompt instance with the added output.
        """

        return Prompt(self.template, self.outputs + [Output(*args, **kwargs)])

    def query(
        self,
        llm_completion: Any = litellm_completion,
        retries: int = 4,
        *args: Any,
        **kwargs: Any,
    ) -> Response:
        """Query an LLM with the prompt and handle retries.

        Args:
            llm_completion (Callable): The LLM completion function.
            retries (int): Number of retry attempts.
            *args: Additional positional arguments for the LLM function.
            **kwargs: Additional keyword arguments for the LLM function.

        Returns:
            Response: The response from the LLM.

        Raises:
            ResponseError: raised when LLM response and Prompt contain incompatible outputs (different number, name or type).
        """

        response_params = inspect.signature(Response.__init__).parameters

        response_kwargs = {k: v for k, v in kwargs.items() if k in response_params}
        llm_completion_kwargs = {
            k: v for k, v in kwargs.items() if k not in response_params
        }

        errors: List[str] = []

        for retry_time in range(retries):

            try:

                if retry_time > 0:
                    warnings.warn("Setting temperature to 1!", RuntimeWarning)
                    llm_completion_kwargs["temperature"] = 1.0

                prompt_text = self.__str__()

                if errors:
                    prompt_text += (
                        "\n\nMake sure to avoid the following errors in the XML:\n"
                    )
                    prompt_text += "\n- ".join(errors)

                raw_response_text = llm_completion(
                    prompt_text, *args, **llm_completion_kwargs
                )
                response = Response(raw_response_text, **response_kwargs)

                new_errors = []

                # Check that expected and responded outputs are compatible
                if len(self.outputs) != len(response):
                    new_errors += [
                        f"Expected {len(self.outputs)} outputs in LLM response, but got {len(response)}"
                    ]
                    errors += new_errors
                    raise ResponseError("\n".join(new_errors))

                new_errors = []
                for index, (defined, responded) in enumerate(
                    zip(self.outputs, response)
                ):
                    if defined.name != responded.name:
                        new_errors += [
                            f"Name for output at position {index} ('{defined.name}') differs from the one provided by LLM ('{responded.name}')\n"
                        ]
                    if defined.type != responded.type:
                        new_errors += [
                            f"Type for output at position {index} ('{defined.type}') differs from the one provided by LLM ('{responded.type}')\n"
                        ]

                if new_errors:
                    errors += new_errors
                    raise ResponseError("\n".join(new_errors))

                break

            except Exception as e:
                warnings.warn(
                    f"WARNING: failed LLM query (try {retry_time} out of {retries}), reason: {str(e)} ({type(e)})",
                    RuntimeWarning,
                )
                if retry_time + 1 == retries:
                    raise e

        return response



# Copyright (c) 2025 foofaraw (GitHub: foofaraw)
# Licensed under the MIT License (see LICENSE file for details).

from .exceptions import (
    ConcatenationError,
    MalformedOutput,
    PromptError,
    ReplacementError,
    ResponseError,
)
from .output import Output
from .prompt import Prompt
from .response import Response

__all__ = [
    "ConcatenationError",
    "ReplacementError",
    "MalformedOutput",
    "PromptError",
    "ResponseError",
    "Output",
    "Prompt",
    "Response",
]



# Copyright (c) 2025 foofaraw (GitHub: foofaraw)
# Licensed under the MIT License (see LICENSE file for details).

from __future__ import (  # Required for forward references in older Python versions
    annotations,
)

from typing import Iterator, List

from .output import Output, xml_to_outputs


class Response:
    """
    A class representing a response containing code blocks.

    Attributes:
        __raw_response_text (str): The raw response text.
        __outputs (List[Output]): The extracted code blocks.
    """

    def __init__(self, response_text: str):
        """
        Initializes a Response object.

        Args:
            response_text (str): The raw response text.
        """
        self.__raw_response_text: str = response_text

        self.__outputs: List[Output] = xml_to_outputs(self.__raw_response_text)

        # Add output contents as member variables in response object
        for output in self.__outputs:
            if output.name and not hasattr(self, output.name):
                setattr(self, output.name, output.content)

    def __str__(self) -> str:
        """Returns the raw response text."""
        return self.__raw_response_text

    def __iter__(self) -> Iterator[Output]:
        """Returns an iterator over the code blocks."""
        return iter(self.__outputs)

    def __len__(self) -> int:
        """Returns the number of code blocks."""
        return len(self.__outputs)

    def __getitem__(self, index: int) -> Output:
        """Returns the code block at the given index."""
        return self.__outputs[index]

    def __contains__(self, name: str) -> bool:
        """
        Checks if a code block with the given language exists.

        Args:
            name (str): The output name.

        Returns:
            bool: True if output with that name exists in response, False otherwise.
        """
        return any(output.name == name for output in self.__outputs)



# Copyright (c) 2025 foofaraw (GitHub: foofaraw)
# Licensed under the MIT License (see LICENSE file for details).

import ast
import re
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from typing import Any, List, Union
from xml.dom import minidom

from lxml import etree

from .exceptions import MalformedOutput, ResponseError


def convert_to_Python_type(value_str: str, type_str: str) -> Any:
    safe_globals = {
        "__builtins__": None,
        "int": int,
        "float": float,
        "str": str,
        "bool": bool,
    }

    if type_str not in safe_globals:
        raise ResponseError(
            f"Tried to create Output with a non-standard basic Python type: '{type_str}'"
        )

    try:
        parsed_type = eval(type_str, safe_globals)
        return parsed_type(ast.literal_eval(value_str))
    except SyntaxError:
        raise MalformedOutput(
            f"Could not cast parameter value '{value_str}' to suggested type '{type_str}'"
        )


@dataclass
class Output:
    """A dataclass representing one output of an LLM query."""

    name: Union[str, None] = None
    description: Union[str, None] = None
    content: Union[str, None] = None
    type: Union[str, None] = None

    def __post_init__(self) -> None:

        if self.name and not self.name.isidentifier():
            raise MalformedOutput(
                f"Invalid output name: '{self.name}'. Must comply with Python identifier format: [a-z_][a-z0-9_-]*"
            )

        if self.type and self.content:
            if self.type != "str":
                self.content = convert_to_Python_type(self.content, self.type)


def outputs_to_xml(outputs: List[Output]) -> str:
    """
    Convert a list of Output objects into an XML string.
    - The fields type, name, and description are XML attributes.
    - The content field is wrapped inside a CDATA section within the <output> element.
    """
    root = etree.Element("outputs")  # Root element

    for out in outputs:
        output_elem = etree.SubElement(root, "output")

        # Set attributes
        if out.name is not None:
            output_elem.set("name", out.name)
        if out.description is not None:
            output_elem.set("description", out.description)
        if out.type is not None:
            output_elem.set("type", out.type)

        # Wrap content in a CDATA section
        if out.content is not None:
            output_elem.text = etree.CDATA(str(out.content))

    # Convert the ElementTree to a string
    xml_str = etree.tostring(root, encoding="unicode").strip()

    # Indent using minidom
    dom = minidom.parseString(xml_str)
    pretty_xml_str = dom.toprettyxml(indent="  ")  # Use 2 spaces for indentation

    # Remove the XML declaration line if present:
    lines = pretty_xml_str.splitlines()
    lines = [line for line in lines if not line.strip().startswith("<?xml")]
    pretty_xml_str = "\n".join(lines)

    return pretty_xml_str


def xml_to_outputs(text: str) -> List[Output]:
    """
    Convert an XML string back into a list of `Output` objects.
    - Reads the `type`, `name`, and `description` from attributes.
    - Uses the text content for `content`.
    """

    # Extracts the XML portion from a larger text using a regular expression.
    # Assumes the XML is enclosed in <Outputs>...</Outputs> tags.
    # Use a regular expression to find the XML portion

    matches = list(re.finditer(r"<outputs>.*?</outputs>", text, re.DOTALL))
    if not matches:
        return []
    xml_string = matches[-1].group(0)

    root = ET.fromstring(xml_string)
    output_list = []

    for elem in root.findall("output"):
        # Get attributes or fallback to None
        name = elem.get("name")
        description = elem.get("description")
        out_type = elem.get("type")

        content = elem.text or None  # if there's no text, store None

        output_list.append(
            Output(
                type=out_type,
                content=content,
                name=name,
                description=description,
            )
        )

    return output_list
